/* Gnome-Streamer
 * Copyright (C) <2001> Richard Boulton <richard-gst@tartarus.org>
 *
 * Based on example.c:
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "esdsink.h"
#include <esd.h>
#include <gst/meta/audioraw.h>

/* elementfactory information */
static GstElementDetails esdsink_details = {
  "Esound audio sink",
  "Sink/Esdsink",
  "Plays audio to an esound server",
  VERSION,
  "Richard Boulton <richard-gst@tartarus.org>",
  "(C) 2001",
};

static gboolean gst_esdsink_open_audio (GstEsdsink *sink);
static void gst_esdsink_close_audio (GstEsdsink *sink);
static GstElementStateReturn gst_esdsink_change_state (GstElement *element);
static gboolean gst_esdsink_sync_parms (GstEsdsink *esdsink);

/* Signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_MUTE,
  ARG_DEPTH,
  ARG_CHANNELS,
  ARG_RATE,
  ARG_HOST,
};

static GstPadFactory sink_factory = {
  "sink",					/* the name of the pads */
  GST_PAD_FACTORY_SINK,				/* type of the pad */
  GST_PAD_FACTORY_ALWAYS,			/* ALWAYS/SOMETIMES */
  GST_PAD_FACTORY_CAPS(
     "esdsink_sink",				/* the name of the caps */
     "audio/raw",				/* the mime type of the caps */
     /* Properties follow: */
     "format",   GST_PROPS_LIST (GST_PROPS_INT (AFMT_S16_LE),
				 GST_PROPS_INT (AFMT_U8)),
     "depth",    GST_PROPS_LIST (GST_PROPS_INT (8), GST_PROPS_INT (16)),
     "rate",     GST_PROPS_INT_RANGE (8000, 96000),
     "channels", GST_PROPS_LIST (GST_PROPS_INT (1), GST_PROPS_INT (2))
  ),
  NULL
};


static void 	gst_esdsink_class_init		(GstEsdsinkClass *klass);
static void 	gst_esdsink_init		(GstEsdsink *esdsink);

static void 	gst_esdsink_chain 		(GstPad *pad, GstBuffer *buf);

static void 	gst_esdsink_set_arg		(GtkObject *object,GtkArg *arg,guint id);
static void 	gst_esdsink_get_arg		(GtkObject *object,GtkArg *arg,guint id);

GstPadTemplate *sink_template;

#define GST_TYPE_ESDSINK_DEPTHS (gst_esdsink_depths_get_type())
static GtkType
gst_esdsink_depths_get_type(void) {
  static GtkType esdsink_depths_type = 0;
  static GtkEnumValue esdsink_depths[] = {
    {8, "8", "8 Bits"},
    {16, "16", "16 Bits"},
    {0, NULL, NULL},
  };
  if (!esdsink_depths_type) {
    esdsink_depths_type = gtk_type_register_enum("GstEsdsinkDepths", esdsink_depths);
  }
  return esdsink_depths_type;
}

#define GST_TYPE_ESDSINK_CHANNELS (gst_esdsink_channels_get_type())
static GtkType
gst_esdsink_channels_get_type(void) {
  static GtkType esdsink_channels_type = 0;
  static GtkEnumValue esdsink_channels[] = {
    {1, "1", "Mono"},
    {2, "2", "Stereo"},
    {0, NULL, NULL},
  };
  if (!esdsink_channels_type) {
    esdsink_channels_type = gtk_type_register_enum("GstEsdsinkChannels", esdsink_channels);
  }
  return esdsink_channels_type;
}


static GstElementClass *parent_class = NULL;
//static guint gst_esdsink_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_esdsink_get_type(void) 
{
  static GtkType esdsink_type = 0;

  if (!esdsink_type) {
    static const GtkTypeInfo esdsink_info = {
      "GstEsdsink",
      sizeof(GstEsdsink),
      sizeof(GstEsdsinkClass),
      (GtkClassInitFunc)gst_esdsink_class_init,
      (GtkObjectInitFunc)gst_esdsink_init,
      (GtkArgSetFunc)gst_esdsink_set_arg,
      (GtkArgGetFunc)gst_esdsink_get_arg,
      (GtkClassInitFunc)NULL,
    };
    esdsink_type = gtk_type_unique(GST_TYPE_ELEMENT,&esdsink_info);
  }
  return esdsink_type;
}

static void 
gst_esdsink_class_init (GstEsdsinkClass *klass) 
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstEsdsink::mute", GTK_TYPE_BOOL,
			  GTK_ARG_READWRITE, ARG_MUTE);
  gtk_object_add_arg_type("GstEsdsink::depth", GST_TYPE_ESDSINK_DEPTHS,
			  GTK_ARG_READWRITE, ARG_DEPTH);
  gtk_object_add_arg_type("GstEsdsink::channels", GST_TYPE_ESDSINK_CHANNELS,
			  GTK_ARG_READWRITE, ARG_CHANNELS);
  gtk_object_add_arg_type("GstEsdsink::frequency", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_RATE);
  gtk_object_add_arg_type("GstEsdsink::host", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_HOST);

  gtkobject_class->set_arg = gst_esdsink_set_arg;  
  gtkobject_class->get_arg = gst_esdsink_get_arg;

  gstelement_class->change_state = gst_esdsink_change_state;
}

static void 
gst_esdsink_init(GstEsdsink *esdsink) 
{
  esdsink->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad(GST_ELEMENT(esdsink), esdsink->sinkpad);
  gst_pad_set_chain_function(esdsink->sinkpad, gst_esdsink_chain);

  esdsink->mute = FALSE;
  esdsink->fd = -1;
  // FIXME: get default from somewhere better than just putting them inline.
  esdsink->format = AFMT_S16_LE;
  esdsink->depth = 16;
  esdsink->channels = 2;
  esdsink->frequency = 44100;
  esdsink->host = NULL;
}

static gboolean
gst_esdsink_sync_parms (GstEsdsink *esdsink)
{
  g_return_val_if_fail (esdsink != NULL, FALSE);
  g_return_val_if_fail (GST_IS_ESDSINK (esdsink), FALSE);

  if (esdsink->fd == -1) return TRUE;

  // Need to set fd to use new parameters: only way to do this is to reopen.
  gst_esdsink_close_audio (esdsink);
  return gst_esdsink_open_audio (esdsink);
}


static void 
gst_esdsink_chain (GstPad *pad, GstBuffer *buf) 
{
  GstEsdsink *esdsink;
  MetaAudioRaw *meta;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  esdsink = GST_ESDSINK(pad->parent);
  g_return_if_fail(esdsink != NULL);
  g_return_if_fail(GST_IS_ESDSINK(esdsink));

  meta = (MetaAudioRaw *)gst_buffer_get_first_meta (buf);
  if (meta != NULL) {
    if ((meta->format != esdsink->format) ||
	(meta->channels != esdsink->channels) ||
	(meta->frequency != esdsink->frequency)) {
      esdsink->format    = meta->format;
      esdsink->depth     = (esdsink->format == AFMT_S16_LE) ? 16 : 8;
      esdsink->channels  = meta->channels;
      esdsink->frequency = meta->frequency;
      if (!gst_esdsink_sync_parms (esdsink)) {
	  GST_DEBUG (0, "esdsink: failed to write buffer due to failure to reopen connection to esound server\n");
	  return;
      }
      GST_DEBUG(0, "esdsink: sound device set to format %d, %d bit depth, %d channels, %dHz\n",
		esdsink->format, esdsink->depth, esdsink->channels, esdsink->frequency);
    }
  }

  if (GST_BUFFER_DATA (buf) != NULL) {
    gst_trace_add_entry(NULL, 0, buf, "esdsink: writing to server");
    if (!esdsink->mute && esdsink->fd >= 0) {
      GST_DEBUG (0, "esdsink: fd=%d data=%p size=%d\n",
		 esdsink->fd, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
      write (esdsink->fd, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
    }
  }
}

static void 
gst_esdsink_set_arg (GtkObject *object,GtkArg *arg,guint id) 
{
  GstEsdsink *esdsink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_ESDSINK(object));
  esdsink = GST_ESDSINK(object);

  switch(id) {
    case ARG_MUTE:
      esdsink->mute = GTK_VALUE_BOOL(*arg);
      break;
    case ARG_DEPTH:
      esdsink->depth = GTK_VALUE_ENUM (*arg);
      gst_esdsink_sync_parms (esdsink);
      break;
    case ARG_CHANNELS:
      esdsink->channels = GTK_VALUE_ENUM (*arg);
      gst_esdsink_sync_parms (esdsink);
      break;
    case ARG_RATE:
      esdsink->frequency = GTK_VALUE_INT (*arg);
      gst_esdsink_sync_parms (esdsink);
      break;
    case ARG_HOST:
      if (esdsink->host != NULL) g_free(esdsink->host);
      if (GTK_VALUE_STRING (*arg) == NULL)
	  esdsink->host = NULL;
      else
	  esdsink->host = g_strdup (GTK_VALUE_STRING (*arg));
      break;
    default:
      break;
  }
}

static void 
gst_esdsink_get_arg (GtkObject *object,GtkArg *arg,guint id) 
{
  GstEsdsink *esdsink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_ESDSINK(object));
  esdsink = GST_ESDSINK(object);

  switch (id) {
    case ARG_MUTE:
      GTK_VALUE_BOOL(*arg) = esdsink->mute;
      break;
    case ARG_DEPTH:
      GTK_VALUE_ENUM (*arg) = esdsink->depth;
      break;
    case ARG_CHANNELS:
      GTK_VALUE_ENUM (*arg) = esdsink->channels;
      break;
    case ARG_RATE:
      GTK_VALUE_INT (*arg) = esdsink->frequency;
      break;
    case ARG_HOST:
      GTK_VALUE_STRING (*arg) = esdsink->host;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

GstPlugin*
plugin_init (GModule *module) 
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("esdsink");
  g_return_val_if_fail (plugin != NULL, NULL);

  factory = gst_elementfactory_new("esdsink", GST_TYPE_ESDSINK, &esdsink_details);
  g_return_val_if_fail (factory != NULL, NULL);

  sink_template = gst_padtemplate_new (&sink_factory);
  gst_elementfactory_add_padtemplate (factory, sink_template);

  gst_plugin_add_factory (plugin, factory);

  return plugin;
}

static gboolean
gst_esdsink_open_audio (GstEsdsink *sink)
{
  // Name used by esound for this connection.
  const char * connname = "gstreamer";
  
  // Bitmap describing audio format.
  esd_format_t esdformat = ESD_STREAM | ESD_PLAY;

  g_return_val_if_fail (sink->fd == -1, FALSE);

  if (sink->depth == 16) esdformat |= ESD_BITS16;
  else if (sink->depth == 8) esdformat |= ESD_BITS8;
  else {
    GST_DEBUG (0, "esdsink: invalid bit depth (%d)\n", sink->depth);
    return FALSE;
  }
  
  if (sink->channels == 2) esdformat |= ESD_STEREO;
  else if (sink->channels == 1) esdformat |= ESD_MONO;
  else {
    GST_DEBUG (0, "esdsink: invalid number of channels (%d)\n", sink->channels);
    return FALSE;
  }

  GST_DEBUG (0, "esdsink: attempting to open connection to esound server\n");
  sink->fd = esd_play_stream_fallback(esdformat, sink->frequency, sink->host, connname);
  if ( sink->fd < 0 ) {
    GST_DEBUG (0, "esdsink: can't open connection to esound server\n");
    return FALSE;
  }

  GST_FLAG_SET (sink, GST_ESDSINK_OPEN);

  return TRUE;
}

static void
gst_esdsink_close_audio (GstEsdsink *sink)
{ 
  if (sink->fd < 0) return;

  close(sink->fd);
  sink->fd = -1;

  GST_FLAG_UNSET (sink, GST_ESDSINK_OPEN);

  g_print("esdsink: closed sound device\n");
}

static GstElementStateReturn
gst_esdsink_change_state (GstElement *element)
{
  g_return_val_if_fail (GST_IS_ESDSINK (element), FALSE);

  /* if going down into NULL state, close the fd if it's open */
  if (GST_STATE_PENDING (element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET (element, GST_ESDSINK_OPEN))
      gst_esdsink_close_audio (GST_ESDSINK (element));
    /* otherwise (READY or higher) we need to open the fd */
  } else {
    if (!GST_FLAG_IS_SET (element, GST_ESDSINK_OPEN)) {
      if (!gst_esdsink_open_audio (GST_ESDSINK (element)))
	return GST_STATE_FAILURE;
    }
  }

  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);
  return GST_STATE_SUCCESS;
}

